import time
from typing                                     import List
from osbot_utils.type_safe.Type_Safe            import Type_Safe
from osbot_utils.type_safe.decorators.type_safe import type_safe
from osbot_aws.apis.Session                     import Session


class Cloud_Front(Type_Safe):

    def client(self):
        return Session().client('cloudfront',region_name='us-east-1')


    def distributions(self):
        response = self.client().list_distributions()       # todo: add support for when not all values are returned
        items = response.get('DistributionList', {}).get('Items', [])
        return items

    def invalidate_path(self, distribution_id: str, paths: str):
        return self.invalidate_paths(distribution_id, [paths])

    @type_safe
    def invalidate_paths(self, distribution_id:str, paths: List[str]):
        kwargs = dict(DistributionId    = distribution_id,
                      InvalidationBatch = { 'Paths'          : { 'Quantity': len(paths)            ,
                                                                 'Items'   : paths                 },
                                            'CallerReference': f'invalidation-{int(time.time()) }' })  # Unique reference
        response = self.client().create_invalidation(**kwargs)
        return response



from osbot_aws.aws.s3.S3__DB_Base import S3__DB_Base


class S3__DB_Base__Disabled(S3__DB_Base):

    def s3(self):
        return None

    def s3_bucket(self):
        return None

    def s3_bucket__temp_data(self):
        return None

    def s3_file_bytes(self, s3_key):
        return None

    def s3_file_contents(self, s3_key):
        return None

    def s3_file_contents_json(self, s3_key):
        return None

    def s3_file_data(self, s3_key):
        return None

    def s3_file_delete(self, s3_key):
        return False

    def s3_file_exists(self, s3_key):
        return False

    def s3_file_info(self, s3_key):
        return {}

    def s3_file_metadata(self, s3_key):
        return {}

    def s3_file_set_metadata(self, s3_key, metadata):
        return False

    def s3_folder_contents(self, folder, return_full_path=False):
        return []

    def s3_folder_files(self, folder='', return_full_path=False, include_sub_folders=False):
        return []

    def s3_folder_list(self, folder='', return_full_path=False):
        return []

    def s3_save_data(self, data, s3_key, metadata=None):
        return False

    def s3_temp_folder__pre_signed_urls_for_object(self, source='NA', reason='NA', who='NA', expiration=3600):
        return {}

    def s3_temp_folder__pre_signed_url(self, s3_bucket, s3_key, operation, expiration=3600):
        return None

    def s3_temp_folder__download_string(self, pre_signed_url):
        return None

    def s3_temp_folder__upload_string(self, pre_signed_url, file_contents):
        return False

    def s3_folder_temp_file_uploads(self):
        return None

    # setup and restore

    def bucket_delete(self):
        return False

    def bucket_delete_all_files(self):
        return False

    def bucket_exists(self):
        return False

    def setup(self):
        return self

    def using_minio(self):
        return False



from osbot_utils.utils.Dev import pprint

from osbot_aws.aws.s3.S3__Key_Generator                 import S3__Key_Generator
from osbot_utils.utils.Json                             import  to_json_str
from osbot_utils.decorators.methods.cache_on_self       import cache_on_self
from osbot_aws.aws.s3.S3                                import S3
from osbot_utils.helpers.flows.models.Flow_Run__Event   import Flow_Run__Event
from osbot_utils.type_safe.Type_Safe                    import Type_Safe
from osbot_utils.helpers.flows.actions.Flow__Events     import flow_events

S3_Folder_NAME__EVENTS_TO_PROCESS = 'events_to_process'
#S3_Folder_NAME__EVENTS_PROCESSED  = 'events_processed'

class Flow_Events__To__S3(Type_Safe):
    s3_bucket        : str = None
    s3_folder        : str = None
    s3_key_generator : S3__Key_Generator

    def __enter__(self):
        return self.start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def add_event_listener(self):
        flow_events.event_listeners.append(self.event_listener)

    def enabled(self):
        return self.s3_bucket is not None                   # todo: add more checks here

    def event_listener(self, flow_event: Flow_Run__Event):
        if self.enabled():
            s3_key        = self.s3_key_for_flow_event(flow_event)
            file_contents = to_json_str(flow_event.json())
            self.s3().file_create_from_string(bucket=self.s3_bucket, key=s3_key, file_contents=file_contents)

    def remove_event_listener(self):
        flow_events.event_listeners.remove(self.event_listener)

    @cache_on_self
    def s3(self):
        return S3()

    def s3_key_for_flow_event(self, flow_event: Flow_Run__Event):
        event_type   = flow_event.event_type.value
        flow_run_id  = flow_event.event_data.flow_run_id
        task_run_id  = flow_event.event_data.task_run_id
        path_elements = self.s3_key_generator.create_path_elements__from_when()
        path_elements.append(flow_run_id)
        if task_run_id:
            path_elements.append(task_run_id)
        path_elements.append(event_type)
        file_id       = flow_event.event_id
        s3_key        = self.s3_key_generator.create_s3_key(path_elements=path_elements, file_id=file_id)
        return s3_key

    def setup(self):
        self.s3_key_generator.root_folder = S3_Folder_NAME__EVENTS_TO_PROCESS
        return self

    def start(self):
        self.setup()
        self.add_event_listener()
        return self

    def stop(self):
        self.remove_event_listener()
        return self


from osbot_utils.utils.Env import not_in_github_action

#todo: refactor this method to osbot_utils
def skip__if_not__in_github_actions():
    import pytest
    if not_in_github_action():
        pytest.skip('For performance reasons only run this test in GitHub Actions')


import os
import warnings

from osbot_aws.aws.lambda_.dependencies.Lambda__Dependency__Inside_Lambda import Lambda__Dependency__Inside_Lambda


# Legacy helper class (will be removed in a future version)

class Lambda__Dependencies():
    def __init__(self, *args, **kwargs):
        warnings.warn("Lambda__Dependencies class is deprecated and will be removed in a future version.", DeprecationWarning,   2,)

    def load_dependencies(self, targets):
        if type(targets) is list:
            for target in targets:
                self.load_dependency(target)
            return
        for target in targets.split(','):
            self.load_dependency(target.strip())

    def load_dependency(self, target):
        if os.getenv('AWS_REGION') is None:
            return
        Lambda__Dependency__Inside_Lambda(package_name=target).load()


# Legacy Static helpers
def load_dependencies(targets):
    Lambda__Dependencies().load_dependencies(targets)

def load_dependency(target):
    Lambda__Dependencies().load_dependency(target)

# def pip_install_dependency(target, target_aws_lambda=True):
#     Lambda__Dependencies().pip_install_dependency(target, target_aws_lambda)
#
# def upload_dependency(target):
#     Lambda__Dependencies().upload_dependency(target)



import re

from osbot_utils.helpers.safe_str.Safe_Str__File__Path import Safe_Str__File__Path

TYPE_SAFE_STR__FILE_PATH__PYTHON_PACKAGE__REGEX      = re.compile(r'[^a-zA-Z0-9_\-./\\ =]')  # Allow alphanumerics, underscores, hyphens, dots, slashes, spaces and equal

class Safe_Str__File__Path__Python_Package(Safe_Str__File__Path):
    regex = TYPE_SAFE_STR__FILE_PATH__PYTHON_PACKAGE__REGEX



from osbot_aws.aws.lambda_                      import boto3__lambda
from osbot_utils.type_safe.decorators.type_safe import type_safe
from osbot_utils.utils.Env                      import load_dotenv
from osbot_aws.apis.shell.Shell_Client          import Shell_Client
from osbot_aws.helpers.Lambda_Layer_Create      import Lambda_Layer_Create
from osbot_aws.OSBot_Setup                      import OSBot_Setup
from osbot_aws.apis.test_helpers.Temp_Aws_Roles import Temp_Aws_Roles
from osbot_aws.helpers.Lambda_Package           import Lambda_Package

# todo: refactor lambda_name to be a Type_Safe variable, but that will clash with the current lambda_name() function (which is used in other projects)
#       stage and the other self.* vars set in the __init__ should also be Type_Safe variables

class Deploy_Lambda:

    def __init__(self, handler, stage=None, lambda_name=None, **kwargs):
        super().__init__(**kwargs)
        load_dotenv()
        self.osbot_setup          = OSBot_Setup()
        self.aws_config           = self.osbot_setup.aws_config
        if type(handler) is str:
            self.handler          = None
            self.module_name      = handler
        else:
            self.handler          = handler
            self.module_name      = handler.__module__

        self.stage                = stage
        self.role_arn             = Temp_Aws_Roles().for_lambda_invocation__role_arn()

        self.package              = self.get_package()
        if lambda_name:                                        # if we have provided a name, use it internally
            self.package.lambda_name     = lambda_name
            self.package.aws_lambda.name = lambda_name
            self.module_name             = lambda_name

        if len(self.lambda_name()) > 64:
            raise ValueError(f'Lambda name too long, it was f{len(self.lambda_name())} and the max is 64: {self.lambda_name()}')

    def __enter__(self                        ): return self
    def __exit__ (self, type, value, traceback): pass

    def add_function_source_code(self):
        root_module_name = self.handler.__module__.split(".").pop(0)
        self.package.add_module(root_module_name)
        return self

    def add_file(self, file_path):
        self.package.add_file(source=file_path)
        return self

    def add_file__boto3__lambda(self):
        self.add_file(boto3__lambda.__file__)
        return self

    def add_folder(self, source, ignore=None):
        self.package.add_folder(source=source, ignore=ignore)
        return self

    def add_layer(self, layer_arn):
        self.package.add_layer(layer_arn)
        return self

    @type_safe
    def add_layers(self, layers_arn: list):
        for layer_arn in layers_arn:
            self.package.add_layer(layer_arn)
        return self

    def add_module(self, module_name):
        self.package.add_module(module_name)
        return self

    def add_modules(self, module_names):
        for module_name in module_names:
            self.package.add_module(module_name)
        return self

    def add_osbot_aws    (self):    self.package.add_osbot_aws    () ; return self
    def add_osbot_browser(self):    self.package.add_osbot_browser() ; return self
    def add_osbot_utils  (self):    self.package.add_osbot_utils  () ; return self
    def add_osbot_elastic(self):    self.package.add_osbot_elastic() ; return self

    def delete(self):
        return self.lambda_function().delete()

    def deploy(self):
        return self.update() == 'Successful'

    def exists(self):
        return self.package.aws_lambda.exists()

    def info(self):
        return self.lambda_function().info()

    def invoke(self, params=None):
        return self.lambda_function().invoke(params)

    def invoke_async(self, params=None):
        return self.lambda_function().invoke_async(params)

    def invoke_return_logs(self, params=None):
        return self.lambda_function().invoke_return_logs(params)

    def files(self):
        return self.package.get_files()

    def function_url(self):
        return self.lambda_function().function_url()

    def get_package(self):
        package = Lambda_Package(self.lambda_name())
        package.aws_lambda.set_s3_bucket(self.osbot_setup.s3_bucket_lambdas)
        package.aws_lambda.set_role(self.role_arn)
        return package

    def lambda_name(self):
        if self.stage is None:
            return self.module_name
        return f"{self.module_name}-{self.stage}"

    def lambda_function(self):
        return self.package.aws_lambda

    def lambda_shell(self):
        return Shell_Client(self.lambda_function())

    def update(self, wait_for_update=True):
        if self.uses_container_image() is False:
            self.add_function_source_code()
            if len(self.package.get_files()) == 0:                       # todo: add this check to the package.update()  method
                raise Exception("There are not files to deploy")
        update_result= self.package.update()
        if update_result.get('status') == 'ok':
            if wait_for_update:
                return self.lambda_function().wait_for_function_update_to_complete()
        return update_result

    def set_container_image(self, image_uri):
        self.package.set_image_uri(image_uri)

    def set_handler(self, handler):
        self.package.set_handler(handler)
        return self

    def set_layers(self, layers):
        self.package.set_layers(layers)

    def set_packages_using_layer(self, packages, skip_layer_creation_if_exists=True):
        layer_name          = f'layer_for__{self.lambda_name()}'
        lambda_layer_create = Lambda_Layer_Create(layer_name)
        lambda_layer_create.add_packages(packages)
        layer_arn           = lambda_layer_create.create(skip_if_exists=skip_layer_creation_if_exists)
        self.lambda_function().add_layer(layer_arn)
        return lambda_layer_create

    def set_env_variable(self, key, value):
        self.package.set_env_variable(key, value)
        return self

    def set_env_variables(self, env_variables):
        self.package.set_env_variables(env_variables)

    def uses_container_image(self):
        return self.package.uses_image_uri()

# legacy methods
Deploy_Lambda.lambda_invoke = Deploy_Lambda.invoke
